/* 
**  mod_webfilter.c -- Apache webfilter module
**  
**  (c) 2003 Dr. Andreas Mueller, Beratung und Entwicklung
*/ 

#include "httpd.h"
#include "http_config.h"
#include "http_protocol.h"
#include "http_request.h"
#include "http_log.h"
#include "ap_config.h"
#include "apr_strings.h"
#include "mod_webfilter_version.h"
#include <gdbm.h>
#include <sys/types.h>
#include <unistd.h>

static char	*mod_webfilter_id = "$Id: mod_webfilter.c,v 1.11 2003/06/29 09:41:30 afm Exp $";

#ifndef WEBFILTER_DEBUG
#define	WEBFILTER_DEBUG	1
#endif /* WEBFILTER_DEBUG */
#if WEBFILTER_DEBUG
#define	webfilter_ap_log_rerror(x, args...) ap_log_rerror(x, ##args)
#else
#define webfilter_ap_log_rerror(args...)
#endif

#define WEBFILTER_TYPE_UNKNOWN	0
#define	WEBFILTER_TYPE_OK	1
#define	WEBFILTER_TYPE_BLACK	2
#define	WEBFILTER_TYPE_WHITE	3

#define	WEBFILTER_MINBUFFER	20480

module AP_MODULE_DECLARE_DATA webfilter_module;

/* shared data */
typedef struct webfilter_stats {
    unsigned long	requests;
    unsigned long	blacklisted;
    unsigned long	whitelisted;
    unsigned long	blacktyped;
    unsigned long	whitetyped;
    unsigned long long	totalsize;
} webfilter_stats;

/* configuration record data type definitions */
typedef struct webfilter_config_rec {
    /* whitelist/blacklist configuration */
    char		*whitelist;	/* db of `allways accept' URLs 	*/
    apr_table_t		*whitelist_categories;	/* need search capability */
    char		*blacklist;	/* db of `allways deny' URLs	*/
    apr_table_t		*blacklist_categories;	/* need search capability */
    char		*whitetypes;
    char		*blacktypes;
    /* tags to filter */
    int			strict;	/* don't accept unlisted hosts */
    apr_size_t		buffersize;
    apr_array_header_t	*tagstofilter;
    apr_table_t		*tagregexs; /* reg exps to use for tag filtering,
					value is ending tag */
    int			maxregex;
    /* shared data for statistics */
    char		*shmname;
    apr_shm_t		*shm;
    webfilter_stats	*stats;
    /* mutex to mediate access to shared data */
    apr_proc_mutex_t	*mutex;
    char		*mutexname;
} webfilter_config_rec;

typedef struct webfilter_ctx {
    apr_bucket_brigade		*bb;
    webfilter_config_rec	*module;
    int				whitelisted;
    int				typecheck;
    unsigned char		*buffer;
    apr_size_t			bytes;	/* current number of bytes in buffer */
    char			*starttag; /*find end tag through regexs table*/
    apr_size_t			startoffset;
    apr_size_t			continueoffset;
    char			*expression;
} webfilter_ctx;

/* The sample content handler */
static int webfilter_table_cmp(void *vr, const char *key, const char *value) {
    request_rec	*r = (request_rec *)vr;
    ap_rprintf(r, "%s ", key);
    return 1;
}

static void webfilter_listtable(request_rec *r, apr_table_t *t) {
    ap_rprintf(r, "    <td>");
    apr_table_do(webfilter_table_cmp, r, t, NULL);
    ap_rprintf(r, "</td>\n");
}

static int webfilter_handler(request_rec *r) {
    webfilter_config_rec	*cfg;

    /* get the configuration record					*/
    cfg = (webfilter_config_rec *)ap_get_module_config(r->per_dir_config,
	&webfilter_module);

    /* make sure we are the right handler				*/
    if (strcmp(r->handler, "webfilter")) {
        return DECLINED;
    }
    r->content_type = "text/html";      

    if (!r->header_only) {
        ap_rputs("<html>\n", r);
	ap_rputs("<head>\n<title>Webfilter Status</title>\n</head>\n", r);
	ap_rputs("<body bgcolor=\"#ffffff\">\n", r);
	ap_rputs("<h1>Webfilter Status</h1>\n", r);
	ap_rputs("<h2>Configuration</h2>\n", r);
	ap_rputs("<table>\n", r);
	ap_rputs("  <tr>\n", r);
	ap_rprintf(r, "    <td>CVS Id:</td><td><i>%s</i></td>\n",
	    mod_webfilter_id);
	ap_rputs("  </tr>\n", r);
	ap_rputs("  <tr>\n", r);
	ap_rprintf(r, "    <td>Version:</td><td>%s</td>\n",
	    mod_webfilter_version);
	ap_rputs("  </tr>\n", r);
	ap_rputs("  <tr>\n", r);
	ap_rprintf(r, "    <td>Blacklist DB:</td><td>%s</td>\n",
	    (cfg->blacklist) ? cfg->blacklist : "not set");
	ap_rputs("  </tr>\n", r);
	if (cfg->blacklist_categories) {
	    ap_rputs("  <tr>\n", r);
	    ap_rputs("    <td valign=\"top\">Categories:</td>\n", r);
	    webfilter_listtable(r, cfg->blacklist_categories);
	    ap_rputs("  </tr>\n", r);
	}
	ap_rputs("  <tr>\n", r);
	ap_rprintf(r, "    <td>Whitelist DB:</td><td>%s</td>\n",
	    (cfg->whitelist) ? cfg->whitelist : "not set");
	ap_rputs("  </tr>\n", r);
	if (cfg->whitelist_categories) {
	    ap_rputs("  <tr>\n", r);
	    ap_rputs("    <td valign=\"top\">Categories:</td>\n", r);
	    webfilter_listtable(r, cfg->whitelist_categories);
	    ap_rputs("  </tr>\n", r);
	}
	ap_rputs("  <tr>\n", r);
	ap_rputs("    <td>Type Whitelist:</td>\n", r);
	ap_rprintf(r, "    <td>%s</td>\n",
	    (cfg->whitetypes) ? cfg->whitetypes : "not set");
	ap_rputs("  </tr>\n", r);
	ap_rputs("  <tr>\n", r);
	ap_rputs("    <td>Type Blacklist:</td>\n", r);
	ap_rprintf(r, "    <td>%s</td>\n",
	    (cfg->blacktypes) ? cfg->blacktypes : "not set");
	ap_rputs("  </tr>\n", r);
	ap_rputs("</table>\n", r);
	ap_rputs("<h2>Statistics</h2>\n", r);
	ap_rputs("<table>\n", r);
	ap_rputs("  <tr>\n", r);
	ap_rprintf(r, "    <td>Requests:</td><td align=\"right\">%lu</td>\n",
	    cfg->stats->requests);
	ap_rputs("  </tr>\n", r);
	if (cfg->blacklist) {
	    ap_rputs("  <tr>\n", r);
	    ap_rputs("    <td>Blacklisted:</td>", r);
	    ap_rprintf(r, "<td align=\"right\">%lu</td>\n",
		cfg->stats->blacklisted);
	    ap_rputs("  </tr>\n", r);
	}
	if (cfg->whitelist) {
	    ap_rputs("  <tr>\n", r);
	    ap_rputs("    <td>Whitelisted:</td>", r);
	    ap_rprintf(r, "<td align=\"right\">%lu</td>\n",
		cfg->stats->whitelisted);
	    ap_rputs("  </tr>\n", r);
	}
	if (cfg->whitetypes) {
	    ap_rputs("  <tr>\n", r);
	    ap_rputs("    <td>Whitelisted Types</td>\n", r);
	    ap_rprintf(r, "    <td align=\"right\">%ld</td>\n",
		cfg->stats->whitetyped);
	    ap_rputs("  </tr>\n", r);
	}
	if (cfg->blacktypes) {
	    ap_rputs("  <tr>\n", r);
	    ap_rputs("    <td>Blacklisted Types</td>\n", r);
	    ap_rprintf(r, "    <td align=\"right\">%ld</td>\n",
		cfg->stats->blacktyped);
	    ap_rputs("  </tr>\n", r);
	}
	ap_rputs("</table>\n", r);
	ap_rputs("<hr />\n", r);
	ap_rputs("&copy; 2003 Dr. Andreas M&uuml;ller, ", r);
	ap_rputs("Beratung und Entwicklung\n", r);
	ap_rputs("</body>\n", r);
	ap_rputs("</html>\n", r);
    }
    return OK;
}

/* the blacklist/whitelist decision is done in the acces control phase */
static int	webfilter_check(const request_rec *r, const char *name,
			GDBM_FILE db,
    const apr_table_t *categories) {
    datum	key, entry;
    int		off;
    char	*catlist;

    webfilter_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r, "[%d] checking %s",
	getpid(), name);
    /* setup the key							*/
    key.dptr = (char *)name; key.dsize = strlen(name) + 1;

    /* get the entry for the key					*/
    entry = gdbm_fetch(db, key);
    if (entry.dptr == NULL) {
	/* no entry found						*/
	webfilter_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r,
	    "[%d] name %s not found", getpid(), name);
	return 0;
    }

    /* scan the entry and check whether any of the categories are in	*/
    /* the table							*/
    webfilter_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r,
	"[%d] checking categories", getpid());
    catlist = (char *)entry.dptr;
    for (off = 0; off < entry.dsize; off += strlen(catlist + off) + 1) {
	webfilter_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r,
	    "[%d] checking entry value [%s]", getpid(), catlist + off);
	if (apr_table_get(categories, catlist + off)) {
	    ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r,
		"[%d] %s list for category %s", getpid(), name, catlist + off);
	    return 1;
	}
    }

    /* if we get to this point, then the entry was not found, so the 	*/
    /* name counts as not listed					*/
    webfilter_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r,
	"[%d] host %s not listed", getpid(), name);
    return 0;
}
static int	webfilter_is_listed(const request_rec *r, const char *hostname,
		const char *database, const apr_table_t *categories) {
    int		rc = 0;
    GDBM_FILE	db;
    const char	*hp;

    ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r, "[%d] checking %s, db %s",
	getpid(), hostname, database);
    /* open the database						*/
    if (NULL == (db = gdbm_open((char *)database, 0, O_RDONLY, 0666, NULL))) {
	ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
	    "[%d] cannot open database %s: %s (%d)", getpid(), database,
	    strerror(errno), errno);
	return 0;
    }
    webfilter_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r, "[%d] db %s open",
	getpid(), database);

    /* for each entry for the hostname we find, check whether it's 	*/
    /* category	 also is in the categories table			*/
    webfilter_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r, "[%d] host %s",
	getpid(), hostname);
    if (webfilter_check(r, hostname, db, categories)) {
	rc = 1;
	goto cleanup;
    }

    /* for each domain name we perform the same check			*/
    hp = ap_strchr(hostname, '.');
    if (NULL == hp)
	goto cleanup;
    for (; hp; hp = ap_strchr(hp + 1, '.')) {
	webfilter_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r, "[%d] domain %s",
	    getpid(), hp);
	if (webfilter_check(r, hp, db, categories)) {
	    rc = 1;
	    goto cleanup;
	}
    }

cleanup:
    /* close database							*/
    webfilter_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r, "[%d] closing db %s",
	getpid(), database);
    gdbm_close(db);
    return rc;
}

static int	webfilter_is_blacklisted(request_rec *r) {
    webfilter_config_rec	*rec;
    ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r, "[%d] checking blacklist",
	getpid());
    rec = (webfilter_config_rec *)ap_get_module_config(r->per_dir_config,
	&webfilter_module);

    /* if the blacklist database is not set, then the URL isn't 	*/
    /* blacklisted							*/
    if (NULL == rec->blacklist)
	return 0;

    /* decide whether the URL is blacklisted				*/
    return webfilter_is_listed(r, r->hostname, rec->blacklist,
	rec->blacklist_categories);
}

static int	webfilter_is_whitelisted(request_rec *r) {
    webfilter_config_rec	*rec;

    ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r, "[%d] checking whitelist",
	getpid());
    rec = (webfilter_config_rec *)ap_get_module_config(r->per_dir_config,
	&webfilter_module);

    /* if the whitelist database is not set, then the URL isn't		*/
    /* blacklisted							*/
    if (NULL == rec->whitelist)
	return 0;

    /* decide whether the URL is whitelisted				*/
    return webfilter_is_listed(r, r->hostname, rec->whitelist,
	rec->whitelist_categories);
}

static apr_status_t	webfilter_access(request_rec *r) {
    webfilter_config_rec	*rec;
    rec = (webfilter_config_rec *)ap_get_module_config(r->per_dir_config,
	&webfilter_module);
    /* first find the hostname for this request				*/
    ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r, "[%d] checking host %s",
	getpid(), r->hostname);
    apr_proc_mutex_lock(rec->mutex);
    rec->stats->requests++;
    apr_proc_mutex_unlock(rec->mutex);

    /* if the name is in the whitelist database, we remember this in	*/
    /* the notes table, but decline the acces decision for some other	*/
    /*  module								*/
    if (webfilter_is_whitelisted(r)) {
	apr_table_set(r->notes, "webfilter:whitelisted", "yes");
	ap_log_rerror(APLOG_MARK, APLOG_INFO, 0, r, "[%d] whitelist accepts %s",
	    getpid(), r->uri);
	apr_proc_mutex_lock(rec->mutex);
	rec->stats->whitelisted++;
	apr_proc_mutex_unlock(rec->mutex);
	webfilter_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r,
	    "[%d] whitelisted (DECLINED)", getpid());
	return DECLINED;
    }

    /* if the name is in the blacklist database, we return a forbid 	*/
    /* reply								*/
    if (webfilter_is_blacklisted(r)) {
	ap_log_rerror(APLOG_MARK, APLOG_INFO, 0, r, "[%d] blacklist rejects %s",
	    getpid(), r->uri);
	apr_proc_mutex_lock(rec->mutex);
	rec->stats->blacklisted++;
	apr_proc_mutex_unlock(rec->mutex);
	/* XXX create an error page containing the category that caused the*/
	/*     blacklisting						*/
	webfilter_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r,
	    "[%d] blacklisted (FORBIDDEN)", getpid());
	return HTTP_FORBIDDEN;
    }

    /* if we get to this point, the URL is listed in neither database,	*/
    /* if the strict flag is set, we will reject the URL		*/
    if (rec->strict) {
	/* XXX create an error page containing the category that caused the*/
	/*     blacklisting						*/
	webfilter_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r,
	    "[%d] not whitelisted (FORBIDDEN)", getpid());
	return HTTP_FORBIDDEN;
    }

    /* by default, we don't know what to do with the URL, so some other	*/
    /* module can decide (but we still reserver the right to work as a 	*/
    /* filter)								*/
    webfilter_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r,
	"[%d] webfilter_access complete, should filter (DECLINED)", getpid());
    return DECLINED;
}

/* check whether the type is acceptable */
static int	webfilter_db_type(request_rec *r, const char *database,
	const char *type) {
    GDBM_FILE	db;
    int		rc = 0;
    datum	key, entry;

    ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r,
	"[%d] checking type %s in databse %s", getpid(), type, database);

    /* handle NULL type case						*/
    if (NULL == type) {
	ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r,
	    "[%d] request for %s has no type", getpid(), r->uri);
	return 0;
    }

    /* open the database						*/
    if (NULL == (db = gdbm_open((char *)database, 0, O_RDONLY, 0666, NULL))) {
	ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
	    "[%d] cannot open database %s: %s (%d)", getpid(), database,
	    strerror(errno), errno);
	return 0;
    }

    /* try to fetch the type from the database				*/
    key.dptr = (char *)type; key.dsize = strlen(type) + 1;
    entry = gdbm_fetch(db, key);
    if (NULL != entry.dptr)
	rc = 1;

    /* close database							*/
    gdbm_close(db);
    return rc;
}

static int	webfilter_check_type(ap_filter_t *f) {
    webfilter_config_rec	*rec;
    webfilter_ctx		*ctx;
    char			*content_type, *p;
    ctx = (webfilter_ctx *)f->ctx;
    rec = ctx->module;

    /* first strip the type of any comments				*/
    if (NULL == f->r->content_type) {
	/* cannot decide without type					*/
	return WEBFILTER_TYPE_OK;
    }
    content_type = (char *)apr_pstrdup(f->r->pool, f->r->content_type);
    if (NULL != (p = strchr(content_type, ';'))) {
	*p = '\0';
    }

    /* first see whether the type is in the whitelist type database	*/
    if (rec->whitetypes) {
	webfilter_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r,
	    "[%d] checking whitelist", getpid());
	if (webfilter_db_type(f->r, rec->whitetypes, content_type)) {
	    apr_proc_mutex_lock(rec->mutex);
	    rec->stats->whitetyped++;
	    apr_proc_mutex_unlock(rec->mutex);
	    return WEBFILTER_TYPE_WHITE;
	}
    }

    /* check the blacklist database					*/
    if (rec->blacktypes) {
	webfilter_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r,
	    "[%d] checking blacklist", getpid());
	if (webfilter_db_type(f->r, rec->blacktypes, content_type)) {
	    apr_proc_mutex_lock(rec->mutex);
	    rec->stats->blacktyped++;
	    apr_proc_mutex_unlock(rec->mutex);
	    return WEBFILTER_TYPE_BLACK;
	}
    }
    return WEBFILTER_TYPE_OK;
}

/* the content filter only takes care of tags */
static void	webfilter_initialize(ap_filter_t *f) {
    webfilter_config_rec	*rec;
    webfilter_ctx		*ctx;
    rec = (webfilter_config_rec *)ap_get_module_config(f->r->per_dir_config,
	&webfilter_module);

    webfilter_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r,
	"[%d] start filtering %s", getpid(), f->r->uri);

    /* initialize the context if it hasn't been so far			*/
    if (NULL == f->ctx) {
	/* allocate memory for the context				*/
	ctx = (webfilter_ctx *)apr_pcalloc(f->r->pool, sizeof(webfilter_ctx));
	f->ctx = ctx;
	ctx->whitelisted = 0;
	ctx->module = rec;
	ctx->typecheck = webfilter_check_type(f);
	webfilter_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r,
	    "[%d] setting up new brigade", getpid());
	ctx->bb = apr_brigade_create(f->r->pool, f->c->bucket_alloc);

	/* read the notes table for the info whether we can bypass	*/
	/* filtering							*/
	webfilter_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r,
	    "[%d] checking whitelisted flag in notes table", getpid());
	if (apr_table_get(f->r->notes, "webfilter:whitelisted"))
	    ctx->whitelisted = 1;

	/* only text/html stuff needs top be tag filtered		*/
	webfilter_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r,
	    "[%d] restrict tag filtering to text types", getpid());
	if (f->r->content_type) {
	    if (strncasecmp(f->r->content_type, "text/html", 9))
		ctx->whitelisted = 1;
	}
	/* XXX should check again later					*/

	/* allocate a buffer for tag filtering				*/
	if (ctx->whitelisted == 0) {
	    if (ctx->module->buffersize < WEBFILTER_MINBUFFER)
		ctx->module->buffersize = WEBFILTER_MINBUFFER;
	    ctx->buffer = apr_palloc(f->r->pool, 1 + 2 * rec->buffersize);
	    *ctx->buffer = '\0';
	    ctx->expression = apr_palloc(f->r->pool, 1 + rec->maxregex);
	}
	ctx->bytes = 0;
	ctx->starttag = NULL;
	ctx->startoffset = -1;
    }
}

#if 0
/* the following two functions are used to compute the max expression length */
static int webfilter_largest_key(void *vr, const char *key,
	const char *value) {
    int	*keysize;
    int	l;
    l = strlen(key);
    if (l > *keysize) *keysize = l;
    return 1;
}

static int	webfilter_maxkeylength(apr_table_t *t) {
    int	l = 0;
    apr_table_do(webfilter_largest_key, &l, t, NULL);
    return l;
}
#endif

#define	MARGIN(rec)	rec->buffersize/2
typedef struct webfilter_match_t {
    char	*subject;
    int		offset, endoff;
    char	*expression;
    int		havematch;
#if WEBFILTER_DEBUG
    request_rec	*r;
#endif
} webfilter_match_t;

static int webfilter_try_match(void *vr, const char *key, const char *value) {
    ap_regex_t		rx;
    ap_regmatch_t	rm;
    webfilter_match_t	*wm = (webfilter_match_t *)vr;
    /* try a match for expression key					*/
    webfilter_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, wm->r, "[%d] trying %s",
	getpid(), key);
    regcomp(&rx, key, AP_REG_ICASE);
    if (0 == regexec(&rx, wm->subject, 1, &rm, 0)) {
	/* a match did occur						*/
	wm->havematch = 1;
	/* if we have a match at the same location, the longer match	*/
	/* takes precedence						*/
	if (rm.rm_so == wm->offset) {
	    /* if the new key is longer, we overwrite the old one	*/
	    if (strlen(wm->expression) < strlen(key)) {
		wm->endoff = rm.rm_eo;
		strcpy(wm->expression, key);
	    }
	}
	/* the following condition means we have an earlier match	*/
	if (rm.rm_so < wm->offset) {
	    wm->offset = rm.rm_so;
	    wm->endoff = rm.rm_eo;
	    strcpy(wm->expression, key);
	}
	webfilter_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, wm->r,
	    "[%d] match for %s at %d", getpid(), key, wm->offset + rm.rm_so);
    }
    regfree(&rx);
    return 1;
}

static int	webfilter_first_match(ap_filter_t *f, int off, char **regex,
    ap_regmatch_t *rm) {
    webfilter_match_t	wm;
    webfilter_ctx	*ctx = (webfilter_ctx *)f->ctx;

    wm.subject = (char *)(ctx->buffer + off);
    wm.offset = strlen(wm.subject);
    wm.expression = ctx->expression;
    wm.havematch = 0;
    ctx = (webfilter_ctx *)f->ctx;
#if WEBFILTER_DEBUG
    wm.r = f->r;
#endif

    /* go through each regex in the table to find the first match	*/
    apr_table_do(webfilter_try_match, &wm, ctx->module->tagregexs, NULL);

    /* if we have a match, return the info about it			*/
    if (wm.havematch) {
	webfilter_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, wm.r,
	    "[%d] first match at %d for %s", getpid(), wm.offset,
	    wm.expression);
	*regex = wm.expression;
	rm->rm_so = wm.offset;
	rm->rm_eo = wm.endoff;
	return 1;
    }
    return 0;
}

static void	webfilter_match(ap_filter_t *f, apr_size_t off) {
    webfilter_ctx	*ctx;
    ap_regex_t		rx;
    ap_regmatch_t	rm;
    ctx = (webfilter_ctx *)f->ctx;

    webfilter_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r,
	"[%d] matching at offset %d", getpid(), off);
    /* if we have a pending search for a end tag, we ignore the off (1)	*/
    if (ctx->startoffset >= 0)
	if (off < ctx->continueoffset) {
	    /* recursion 						*/
	    webfilter_match(f, ctx->continueoffset);
	    return;
	}

    /* only match if enough data is left				*/
    if (off > ctx->module->buffersize)
	return;

    /* if we have a pending open tag, we look for the closing tag	*/
    /* by (1) above, we can be sure that off = ctx->startoffset		*/
    if (ctx->starttag) {
	/* first case: pending start tag, waiting to be closed		*/
	const char	*endrx;
	int		limit = ctx->module->buffersize;
	int		from = ctx->continueoffset;

	webfilter_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r,
	    "[%d] pending tag %s", getpid(), ctx->starttag);
	/* find the corresponding end tag regex				*/
	endrx = apr_table_get(ctx->module->tagregexs, ctx->starttag);
	webfilter_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r,
	    "[%d] waiting for closing tag %s", getpid(), endrx);

	/* look for the thing						*/
	regcomp(&rx, endrx, AP_REG_ICASE);
	webfilter_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r,
	    "[%d] regex %s compiled", getpid(), endrx);
	if (0 == regexec(&rx, (char *)(ctx->buffer + off), 1, &rm, 0)) {
	    /* remember where the match ends, and forget the start tag	*/
	    limit = off + rm.rm_eo;
	    ctx->starttag = NULL;
	    ctx->startoffset = -1;
	    webfilter_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r,
		"[%d] padding from %d to %d", getpid(), from, limit);
	    memset(ctx->buffer + from, ' ', limit - from);

	    /* rematch							*/
	    webfilter_match(f, limit);
	} else {
	    webfilter_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r,
		"[%d] padding from %d to middle of buffer", getpid(), from);
	    memset(ctx->buffer + from, ' ', limit - from);
	}
	regfree(&rx);

    } else {
	/* second case: no tag pending					*/
	char		*regex;
	const char	*close;
	/* find first match						*/
	webfilter_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r,
	    "[%d] looking for first match after %d", getpid(), off);
	if (webfilter_first_match(f, off, &regex, &rm)) {
	    webfilter_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r,
		"[%d] found match with %s at %d", getpid(), regex,
		off + rm.rm_so);
	    /* we have found a match					*/
	    close = apr_table_get(ctx->module->tagregexs, regex);
	    if (strlen(close) > 0) {
		/* there must also be a closing tag, so we initiate the	*/
		/* search for it					*/
		ctx->starttag = (char *)apr_pstrdup(f->r->pool, regex);
		ctx->startoffset = off + rm.rm_so;
		ctx->continueoffset = off + rm.rm_eo;
	    } else {
		webfilter_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r,
		    "[%d] found a simple match, overwriting from %d to %d",
		    getpid(), off + rm.rm_so, off + rm.rm_eo);
		/* we overwrite the complete tag with blanks		*/
	    }
	    memset(ctx->buffer + off + rm.rm_so, ' ', rm.rm_eo - rm.rm_so);

	    /* try again from the end of the last match			*/
	    webfilter_match(f, off + rm.rm_eo);
	}
    }
}

static apr_status_t webfilter_filter(ap_filter_t *f, apr_bucket_brigade *bb) {
    webfilter_ctx	*ctx;
    int			complete = 0, doflush = 0;
    const char		*str = NULL;
    apr_size_t		len = 0, remain;
    apr_bucket		*e;

    /* make sure the context is properly initialized			*/
    webfilter_initialize(f);
    ctx = (webfilter_ctx *)f->ctx;

    /* if we are set to strict, then this function is only called for	*/
    /* whitelisted URLs. As we want to filter in this case, we just	*/
    /* skip the whitelist test						*/
    if (!ctx->module->strict) {
	/* if the URL is whitelisted, just pass the content to the next */
	/* filter 							*/
	if ((ctx->whitelisted) || (ctx->typecheck == WEBFILTER_TYPE_WHITE)) {
	    return ap_pass_brigade(f->next, bb);
	}
    }

    /* if the type is blacklisted, we should as soon as possible send	*/
    /* an error bucket to the client, because the headers probably 	*/
    /* haven't been sent yet						*/
    if (ctx->typecheck == WEBFILTER_TYPE_BLACK) {
	/* XXX format an error page for this situation and return a 	*/
	/*     FORBIDDEN message					*/
	APR_BRIGADE_INSERT_TAIL(ctx->bb,
	    ap_bucket_error_create(HTTP_FORBIDDEN,
		"access denied: blacklisted datatype", f->r->pool,
		f->c->bucket_alloc));
	return ap_pass_brigade(f->next, ctx->bb);
    }

    /* read as many buckets as we can from bb into the buffer		*/
    webfilter_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r,
	"[%d] checking brigade", getpid());
    for (e = APR_BRIGADE_FIRST(bb);
         e != APR_BRIGADE_SENTINEL(bb);
         e = APR_BUCKET_NEXT(e)) {
	/* ignore flush buckets						*/
	if (APR_BUCKET_IS_FLUSH(e)) {
	    webfilter_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r,
		"[%d] removing flush bucket", getpid());
	    APR_BUCKET_REMOVE(e);
	    doflush = 1;
	    continue;
	}
	if (APR_BUCKET_IS_EOS(e)) {
	    webfilter_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r,
		"[%d] removing eos bucket", getpid());
	    APR_BUCKET_REMOVE(e);
	    complete = 1;
	    break;
	}
	/* if we get here, we should try to read the data into the 	*/
	/* buffer							*/
	apr_bucket_read(e, &str, &len, APR_NONBLOCK_READ);
	webfilter_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r,
	    "[%d] got %d bytes from bucket", getpid(), len);
	
	remain = (2 * ctx->module->buffersize) - ctx->bytes;
	if (len > remain) {
	    ap_log_rerror(APLOG_MARK, APLOG_CRIT, 0, f->r,
		"[%d] more data (%d) in bucket than space in buffer (%d)",
		getpid(), len, remain);
	    return APR_EGENERAL;
	}
	webfilter_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r,
	    "[%d] appending data to existing buffer", getpid());
	memcpy(ctx->buffer + ctx->bytes, str, len); ctx->bytes += len;
	ctx->buffer[ctx->bytes] = '\0';	/* make sure string is 0 terminated */
	if (ctx->bytes > ctx->module->buffersize + MARGIN(ctx->module)) {
	    webfilter_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r,
		"[%d] performing match with %d bytes", getpid(), ctx->bytes);
	    webfilter_match(f, 0);
	    /* flush buffer to client					*/
	    webfilter_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r,
		"[%d] sending buffer contents (%d bytes) to client", getpid(),
		ctx->module->buffersize);
	    ap_fwrite(f->next, ctx->bb, (char *)ctx->buffer,
		ctx->module->buffersize);
	    webfilter_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r,
		"[%d] moving %d bytes to front, from %x to %x", getpid(),
		ctx->bytes - ctx->module->buffersize + 1,
		(unsigned int)(ctx->buffer + ctx->module->buffersize),
		(unsigned int)(ctx->buffer));
	    memmove(ctx->buffer, ctx->buffer + ctx->module->buffersize,
		ctx->bytes - ctx->module->buffersize + 1);
		/* move the terminating 0 as well!			*/
	    ctx->bytes -= ctx->module->buffersize;
	    webfilter_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r,
		"[%d] buffer now contains %d bytes", getpid(), ctx->bytes);
	    if ((ctx->startoffset >= 0)
		&& (ctx->startoffset < ctx->module->buffersize)) {
		ctx->startoffset = 0;
		ctx->continueoffset -= ctx->module->buffersize;
		if (ctx->continueoffset < 0)
		    ctx->continueoffset = 0;
	    }
	}
    }

    /* if we have all the content, match one last time and flush 	*/
    /* everything							*/
    if (complete) {
	webfilter_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r,
	    "[%d] closing", getpid());
	webfilter_match(f, 0);
	webfilter_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r,
	    "[%d] sending last %d bytes to client", getpid(), ctx->bytes);
	ap_fwrite(f->next, ctx->bb, (char *)ctx->buffer, ctx->bytes);
	APR_BRIGADE_INSERT_TAIL(ctx->bb,
	    apr_bucket_eos_create(f->c->bucket_alloc));
    }
    if (doflush) {
	APR_BRIGADE_INSERT_TAIL(ctx->bb,
	    apr_bucket_flush_create(f->c->bucket_alloc));
    }
    ap_pass_brigade(f->next, ctx->bb);

    /* apr_brigade_cleanup(bb);						*/
    return APR_SUCCESS;
}

/* register the various phases with the server */
static void webfilter_register_hooks(apr_pool_t *p) {
    /* we don't really need the predecessor and successor arrays, but	*/
    /* if you have special needs in your application, they are already	*/
    /* here								*/
    static const char	*aszPre[] = { NULL };
    static const char	*aszSucc[] = { NULL };

    /* make sure the webfilter access check happens fairly early	*/
    ap_hook_access_checker(webfilter_access, aszPre, aszSucc, APR_HOOK_FIRST);

    /* we aren't picky about when the content handler is called		*/
    ap_hook_handler(webfilter_handler, NULL, NULL, APR_HOOK_MIDDLE);

    /* the webfilter filter handler applies to content			*/
    ap_register_output_filter("WEBFILTER", webfilter_filter, NULL,
	AP_FTYPE_CONTENT_SET);
}

/* configuration methods */
static void	*webfilter_create_dir_config(apr_pool_t *p, char *d) {
    apr_status_t		rc;
    webfilter_config_rec	*cfg;
    cfg = (webfilter_config_rec *)apr_pcalloc(p, sizeof(webfilter_config_rec));
    cfg->whitelist = NULL;
    cfg->whitelist_categories = apr_table_make(p, 0);
    cfg->blacklist = NULL;
    cfg->blacklist_categories = apr_table_make(p, 0);
    cfg->blacktypes = NULL;
    cfg->whitetypes = NULL;
    cfg->tagstofilter = apr_array_make(p, 0, sizeof(char *));
    cfg->tagregexs = apr_table_make(p, 0);
    cfg->maxregex = 0;
    cfg->strict = 0;

    /* initialize mutex and shared memory */
    if (APR_SUCCESS != apr_proc_mutex_create(&cfg->mutex, cfg->mutexname,
	APR_LOCK_DEFAULT, p)) {
	fprintf(stderr, "%s:%d: cannot create mutex %s", __FILE__, __LINE__,
	    cfg->mutexname);
    }

    /* create shared memory (mutex is required for this) */
    apr_proc_mutex_lock(cfg->mutex);
    rc = apr_shm_attach(&cfg->shm, cfg->shmname, p);
    if (APR_SUCCESS != rc) {
	/* since we failed to attach to the shared memory, try to 	*/
	/* create it							*/
	rc = apr_shm_create(&cfg->shm, sizeof(webfilter_stats),
	    cfg->shmname, p);
	if (APR_SUCCESS != rc) {
	    fprintf(stderr, "%s:%d: cannot create shared memory %s\n",
		__FILE__, __LINE__, cfg->shmname);
	} else {
	    /* initialize the shared memory				*/
	    cfg->stats = (webfilter_stats *)apr_shm_baseaddr_get(cfg->shm);
	    memset(cfg->stats, 0, sizeof(webfilter_stats));
	}
    } else {
	/* initialize the pointer to the shared structure, but dont 	*/
	/* initialize							*/
	cfg->stats = (webfilter_stats *)apr_shm_baseaddr_get(cfg->shm);
    }
    apr_proc_mutex_unlock(cfg->mutex);

    /* the API expects that we return a pointer to the new 		*/
    /* configuration							*/
    return cfg;
}

static void	*webfilter_merge_dir_config(apr_pool_t *p, void *pp, void *cp) {
    int				i;
    webfilter_config_rec	*parent = (webfilter_config_rec *)pp;
    webfilter_config_rec	*child = (webfilter_config_rec *)cp;

    if (child->whitelist != NULL) child->whitelist = parent->whitelist;
    if (child->blacklist != NULL) child->blacklist = parent->blacklist;
    if (child->whitetypes != NULL) child->whitetypes = parent->whitetypes;
    if (child->blacktypes != NULL) child->blacktypes = parent->blacktypes;
    if (child->mutexname != NULL) child->mutexname = parent->mutexname;
    if (child->shmname != NULL) child->shmname = parent->shmname;
    child->strict = parent->strict;

    if (apr_is_empty_table(child->blacklist_categories))
	apr_table_overlay(p, parent->blacklist_categories,
	    child->blacklist_categories);
    if (apr_is_empty_table(child->whitelist_categories))
	apr_table_overlay(p, parent->whitelist_categories,
	    child->whitelist_categories);
    if (child->tagstofilter->nelts == 0) {
	for (i = 0; i < parent->tagstofilter->nelts; i++) {
	    char	**tag;
	    tag = apr_array_push(child->tagstofilter);
	    *tag = (char *)apr_pstrdup(p,
		((char **)parent->tagstofilter->elts)[i]);
	}
    }
    if (apr_is_empty_table(child->tagregexs)) {
	apr_table_overlay(p, parent->tagregexs,
	    child->tagregexs);
	child->maxregex = parent->maxregex;
    }

    return cp;
}

static const char	*webfilter_add_whitelist_category(cmd_parms *parms,
	void *mconfig, const char *arg) {
    webfilter_config_rec	*rec = (webfilter_config_rec *)mconfig;
    apr_table_set(rec->whitelist_categories, arg, arg);
    return NULL;
}

static const char	*webfilter_add_blacklist_category(cmd_parms *parms,
	void *mconfig, const char *arg) {
    webfilter_config_rec	*rec = (webfilter_config_rec *)mconfig;
    apr_table_set(rec->blacklist_categories, arg, arg);
    return NULL;
}

static const char	*webfilter_add_regex(cmd_parms *parms,
	void *mconfig, const char *rxstart, const char *rxend) {
    int				l;
    webfilter_config_rec	*rec = (webfilter_config_rec *)mconfig;
    if (rxend)
	apr_table_set(rec->tagregexs, rxstart, rxend);
    else
	apr_table_set(rec->tagregexs, rxstart, "");
    l = strlen(rxstart);
    if (l > rec->maxregex)
	rec->maxregex = l;
    return NULL;
}

static const char	*webfilter_add_tag(cmd_parms *parms,
	void *mconfig, const char *arg) {
    char	**tag;
    char	*start, *end;
    webfilter_config_rec	*rec = (webfilter_config_rec *)mconfig;
    /* add the tag							*/
    tag = (char **)apr_array_push(rec->tagstofilter);
    *tag = (char *)apr_pstrdup(rec->tagstofilter->pool, arg);
    /* add a pair of regular expressions to the tagregexs table for	*/
    /* <tag ..> ... </tag>						*/
    start = apr_psprintf(parms->server->process->pconf, "<%s[^>]*>", arg);
    end = apr_psprintf(parms->server->process->pconf, "</%s[^>]*>", arg);
    webfilter_add_regex(parms, mconfig, start, end);
    /* add expression for <tag ... />					*/
    start = apr_psprintf(parms->server->process->pconf, "<%s[^>]*/>", arg);
    webfilter_add_regex(parms, mconfig, start, NULL);
    return NULL;
}

static const command_rec	webfilter_cmds[] = {
    AP_INIT_TAKE1(
	"WebfilterBlacktypes", ap_set_string_slot,
	(void *)APR_OFFSETOF(webfilter_config_rec, blacktypes),
	RSRC_CONF,
	"Directory containing blacklist database for content types"
    ),
    AP_INIT_TAKE1(
	"WebfilterWhitetypes", ap_set_string_slot,
	(void *)APR_OFFSETOF(webfilter_config_rec, whitetypes),
	RSRC_CONF,
	"Directory containing whitelist database for content types"
    ),
    AP_INIT_TAKE1(
	"WebfilterBlacklist", ap_set_string_slot,
	(void *)APR_OFFSETOF(webfilter_config_rec, blacklist),
	RSRC_CONF,
	"Directory containing blacklist database"
    ),
    AP_INIT_TAKE1(
	"WebfilterWhitelist", ap_set_string_slot,
	(void *)APR_OFFSETOF(webfilter_config_rec, whitelist),
	RSRC_CONF,
	"Directory containing whitelist database"
    ),
    AP_INIT_TAKE1(
	"WebfilterShmname", ap_set_string_slot,
	(void *)APR_OFFSETOF(webfilter_config_rec, shmname),
	RSRC_CONF,
	"Name of shared memory segment"
    ),
    AP_INIT_TAKE1(
	"WebfilterMutexname", ap_set_string_slot,
	(void *)APR_OFFSETOF(webfilter_config_rec, mutexname),    
	RSRC_CONF,
	"Name of mutex that mediates access to shared memory"     
    ),
    AP_INIT_ITERATE(
	"WebfilterBlacklistCategories", webfilter_add_blacklist_category,
	(void *)APR_OFFSETOF(webfilter_config_rec, blacklist_categories),
	RSRC_CONF,
	"a list of categories to blacklist"
    ),
    AP_INIT_ITERATE(
	"WebfilterWhitelistCategories", webfilter_add_whitelist_category,
	(void *)APR_OFFSETOF(webfilter_config_rec, whitelist_categories),
	RSRC_CONF,
	"a list of categories to whitelist"
    ),
    AP_INIT_ITERATE(
	"WebfilterTags", webfilter_add_tag,
	(void *)APR_OFFSETOF(webfilter_config_rec, tagstofilter),
	RSRC_CONF,
	"a list of tags to filter"
    ),
    AP_INIT_TAKE12(
	"WebfilterRegexp", webfilter_add_regex,
	(void *)APR_OFFSETOF(webfilter_config_rec, tagregexs),
	RSRC_CONF,
	"a regular expression (pair) to replace by blanks"
    ),
    AP_INIT_TAKE1(
	"WebfilterBuffersize", ap_set_int_slot,
	(void *)APR_OFFSETOF(webfilter_config_rec, buffersize),    
	RSRC_CONF,
	"size of buffer to keep for scanning before sending data to client"
    ),
    AP_INIT_TAKE1(
	"WebfilterStrict", ap_set_flag_slot,
	(void *)APR_OFFSETOF(webfilter_config_rec, strict),
	RSRC_CONF,
	"set to on if allowing _only_ whitelist hosts, and filtering them"
    ),
    { NULL }
};

/* Dispatch list for API hooks */
module AP_MODULE_DECLARE_DATA webfilter_module = {
    STANDARD20_MODULE_STUFF, 
    webfilter_create_dir_config,
                           /* create per-dir    config structures */
    webfilter_merge_dir_config,
                           /* merge  per-dir    config structures */
    NULL,                  /* create per-server config structures */
    NULL,                  /* merge  per-server config structures */
    webfilter_cmds,        /* table of config file commands       */
    webfilter_register_hooks  /* register hooks                      */
};

