/*
 * webfilter_create.c -- create a database and read the information from the
 *                       standard input into the database
 *
 * (c) 2003 Dr. Andreas Mueller, Beratung und Entwicklung
 */
#if HAVE_CONFIG_H
#include <config.h>
#endif
#include <stdlib.h>
#include <stdio.h>
#include <gdbm.h>
#include <fcntl.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>

#define	LINELENGTH	10240

int	debug = 0;
int	cont = 0;
int	xc = EXIT_SUCCESS;
int	flags = GDBM_INSERT;

#if HAVE_STRSEP
/* we don't need anything, as strsep is implemented on our platform	*/
#else
char	*our_strsep(char **stringp, const char *delim) {
	char	*p, *result = *stringp;

	/* check arguments						*/
	if (result == NULL) return NULL;

	/* perform the search for a delimiter character			*/
	if (NULL == (p = strpbrk(*stringp, delim))) {
		*stringp = NULL;
		return result;
	}

	/* at this point we know that a delimiter character has been	*/
	/* found at p							*/
	*p = '\0';
	*stringp = p + 1;
	return result;
}
#define	strsep	our_strsep
#endif

static void	add_line(GDBM_FILE db, char *line) {
	char	wl[LINELENGTH];
	char	edata[LINELENGTH];
	char	*w, *p, *hostname, *comment = NULL;
	datum	key, entry;
	int	off, l, keysize;

	/* copy the line to the work buffer				*/
	memcpy(wl, line, sizeof(wl));

	/* first check for a comment					*/
	if (NULL != (comment = strchr(wl, '#'))) {
		*comment = '\0';
		comment++;
		while (isspace(*comment)) {
			*comment = '\0';
			comment++;
		}
		if (debug)
			fprintf(stderr, "%s:%d: found a comment: [%s]\n",
				__FILE__, __LINE__, comment);
	}

	/* start tokenizing the line					*/
	p = wl;
	hostname = strsep(&p, " \t\n");
	if (debug)
		fprintf(stderr, "%s:%d: hostname [%s]\n", __FILE__, __LINE__,
			hostname);
	keysize = strlen(hostname) + 1;

	/* add the comment						*/
	if (comment) {
		key.dsize = keysize + 1;
		key.dptr = (char *)alloca(key.dsize);
		strcpy((char *)key.dptr, hostname);
		strcat((char *)key.dptr, "#");
		entry.dptr = comment;
		entry.dsize = strlen(comment) + 1;
		if (gdbm_store(db, key, entry, flags)) {
			fprintf(stderr, "%s:%d: cannot %s comment entry "
				"for %s\n", __FILE__, __LINE__,
				(flags == GDBM_INSERT) ? "insert" : "replace",
				hostname);
		}
	}

	/* tokenize the line into individual options			*/
	off = 0;
	for (w = strsep(&p, " \t\n"); w; w = strsep(&p, " \t\n")) {
		if ((l = strlen(w)) > 0) {
			if (debug)
				fprintf(stderr, "%s:%d: entry [%s]\n",
					__FILE__, __LINE__, w);
			memcpy(edata + off, w, l + 1);
			off += l + 1;
		}
	}

	/* add the entry						*/
	key.dptr = hostname;
	key.dsize = keysize;
	entry.dptr = edata;
	entry.dsize = off;
	if (gdbm_store(db, key, entry, flags)) {
		fprintf(stderr, "%s:%d: cannot %s entry for %s\n",
			__FILE__, __LINE__,
			(flags == GDBM_INSERT) ? "insert" : "replace", hostname);
		if (!cont)
			exit(EXIT_FAILURE);
		else
			xc = EXIT_FAILURE;
	}

}

int	main(int argc, char *argv[]) {
	int		c, truncate = 0;
	char		*dbname;
	GDBM_FILE	db;
	char		line[LINELENGTH];

	/* parse command line						*/
	while (EOF != (c = getopt(argc, argv, "dctr")))
		switch (c) {
		case 'd':
			debug++;
			break;
		case 'c':
			cont = 1;
			break;
		case 't':
			truncate = 1;
			break;
		case 'r':
			flags = GDBM_REPLACE;
			break;
		}

	/* next argument must be database name				*/
	if ((argc - 1) < optind) {
		fprintf(stderr, "%s: must specify database name\n", argv[0]);
		exit(EXIT_FAILURE);
	}
	dbname = argv[optind];

	/* create and truncate database					*/
	if (debug)
		fprintf(stderr, "%s:%d: trying to open database %s\n",
			__FILE__, __LINE__, dbname);
	if (NULL == (db = gdbm_open(dbname, 0, O_RDWR | O_CREAT |
		((truncate) ? O_TRUNC : 0), 0666, NULL))) {
		fprintf(stderr, "%s: cannot create database %s: %s (%d)\n",
			argv[0], dbname, strerror(errno), errno);
		exit(EXIT_FAILURE);
	}

	/* start reading stuff from standard input			*/
	while (fgets(line, sizeof(line), stdin)) {
		if (debug)
			fprintf(stderr, "%s:%d: adding line %s", __FILE__,
				__LINE__, line);
		/* parse the line and add contents to database		*/
		line[strlen(line) - 1] = '\0';
		add_line(db, line);
	}

	/* close database						*/
	if (debug)
		fprintf(stderr, "%s:%d: closing database %s\n", __FILE__,
			__LINE__, dbname);
	gdbm_close(db);

	exit(xc);
}
