/*
 * webfilter_types.c
 *
 * (c) 2003 Dr. Andreas Mueller, Beratung und Entwicklung
 *
 * $Id: webfilter_types.c,v 1.6 2003/06/27 11:04:28 afm Exp $
 */
#if HAVE_CONFIG_H
#include <config.h>
#endif
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <fcntl.h>
#include <ctype.h>
#include <gdbm.h>
#include <unistd.h>

int	debug = 0;

static void	usage(char *progname) {
	printf("usage: %s typedb [ -C comment ] [ -a type ] [ -r type ] "
		"[ -l ] [ -f file ]\n", progname);
}

static void	type_add(GDBM_FILE db, char *type, char *comment) {
	datum	key, entry;
	key.dptr = type; key.dsize = strlen(type) + 1;
	if (comment) {
		entry.dptr = comment; entry.dsize = strlen(comment) + 1;
		if (debug)
			fprintf(stderr, "%s:%d: type [%s], comment [%s]\n",
				__FILE__, __LINE__, type, comment);
	} else {
		entry.dptr = type; entry.dsize = strlen(type) + 1;
		if (debug)
			fprintf(stderr, "%s:%d: type [%s]\n",
				__FILE__, __LINE__, type);
	}
	if (gdbm_store(db, key, entry, GDBM_REPLACE))
		fprintf(stderr, "adding type '%s' failed\n", type);
}

static void	type_delete(GDBM_FILE db, char *type) {
	datum	key;
	key.dptr = type;
	key.dsize = strlen(type) + 1;
	if (gdbm_delete(db, key))
		fprintf(stderr, "deleting type '%s' failed\n", type);
}

#define	LINELENGTH	1024
static void	type_file(GDBM_FILE db, char *filename) {
	char	line[LINELENGTH];
	char	*comment, *p;
	FILE	*fp = fopen(filename, "r");
	if (fp == NULL) {
		fprintf(stderr, "cannot open file %s: %s\n", filename,
			strerror(errno));
		return;
	}
	/* read lines from the file and add each line (minus \n) as a 	*/
	/* type								*/
	while (fgets(line, sizeof(line), fp)) {
		/* kill newline a the end of the line			*/
		line[strlen(line) - 1] = '\0';
		if (debug)
			fprintf(stderr, "%s:%d: analyzing line [%s]\n",
				__FILE__, __LINE__, line);
		/* separate comment					*/
		if (NULL != (comment = strchr(line, '#'))) {
			if (debug)
				fprintf(stderr, "%s:%d: comment detected: [%s]\n",
					__FILE__, __LINE__, comment);
			*comment = '\0';
			comment++;
			/* kill leading whitespace in the comment	*/
			while (isspace(*comment)) {
				*comment = '\0'; comment++;
			}
		}
		/* kill whitespace at the end of the type		*/
		if (NULL != (p = strpbrk(line, " \t")))
			*p = '\0';
		/* add the type to the database				*/
		type_add(db, line, comment);
	}
	fclose(fp);
}

static void	type_list(GDBM_FILE db) {
	datum	key, entry;
	for (key = gdbm_firstkey(db); key.dptr != NULL; key = gdbm_nextkey(db, key)) {
		entry = gdbm_fetch(db, key);
		if (strcmp((char *)key.dptr, (char *)entry.dptr)) {
			printf("%s#%s\n", (char *)key.dptr, (char *)entry.dptr);
		} else {
			printf("%s\n", (char *)key.dptr);
		}
	}
}

int	main(int argc, char *argv[]) {
	int		c;
	GDBM_FILE	db;
	char		*dbname;
	char		*comment = NULL;

	/* the first argument is always the database, so we shop for 	*/
	/* that								*/
	if (argc < 2) {
		fprintf(stderr, "%s: database argument required\n", argv[0]);
		exit(EXIT_FAILURE);
	}
	dbname = argv[1];
	if (NULL == (db = gdbm_open(dbname, 0, O_RDWR | O_CREAT, 0666, NULL))) {
		fprintf(stderr, "%s: cannot open database %s: %s\n", argv[0],
			dbname, strerror(errno));
		exit(EXIT_FAILURE);
	}

	/* now that we have isolated the database, we can start looking	*/
	/* at the rest of the command line parameters			*/
	while (EOF != (c = getopt(argc - 1, argv + 1, "da:r:f:lC:?h")))
		switch (c) {
		case 'C':
			comment = optarg;
			break;
		case 'd':
			debug++;
			break;
		case 'a':
			type_add(db, optarg, comment);
			comment = NULL; /* make sure a comment is valid only */
					/* for one add operation	     */
			break;
		case 'r':
			type_delete(db, optarg);
			break;
		case 'f':
			type_file(db, optarg);
			break;
		case 'l':
			type_list(db);
			break;	
		case 'h':
		case '?':
			usage(argv[0]);
			exit(EXIT_SUCCESS);
		}

	/* close the database						*/
	gdbm_close(db);

	exit(EXIT_SUCCESS);
}
